///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoProcException.cc
// --------------------
// Cego procedure exception container class
//      
// Design and Implementation by Bjoern Lemke
//     
// (C)opyright 2000-2013 Bjoern Lemke
//
// IMPLEMENTATION MODULE
//
// Class: CegoProcException
// 
// Description: 
//
// Status: QG-2.6
//
///////////////////////////////////////////////////////////////////////////////

// cego includes
#include "CegoDefs.h"
#include "CegoProcException.h"

CegoProcException::CegoProcException(CegoException excep, CegoProcBlock *pBlock, CegoProcBlock *pParentBlock) : CegoProcStmt(pParentBlock)
{
    _excep = excep;
    _pBlock = pBlock;
}

CegoProcException::~CegoProcException()
{
}
    
CegoException CegoProcException::getException() const
{
    return _excep;
}

CegoProcBlock* CegoProcException::getBlock()
{
    return _pBlock;
}

CegoException CegoProcException::execute()
{
    return NONE_EXCEP;
}

CegoProcException& CegoProcException::operator = ( const CegoProcException& pe)
{
    _excep = pe._excep;
    _pBlock = pe._pBlock;
    return (*this);
}

bool CegoProcException::operator == ( const CegoProcException& pe)
{
    if ( _excep == pe._excep )
	return true;
    return false;
}

Chain CegoProcException::toChain(const Chain& indent) const
{
    Chain s = indent;

    if ( _excep == INVALID_OBJECT_EXCEP )
    {
	s += Chain(INVALID_OBJECT_EXCEP_ID);
    }
    else if ( _excep == INVALID_ATTRIBUTE_EXCEP )
    {
	s += Chain(INVALID_ATTRIBUTE_EXCEP_ID);
    }
    else if ( _excep == COREOP_EXCEP ) 
    {
	s += Chain(COREOP_EXCEP_ID);
    }
    else if ( _excep == ANY_EXCEP ) 
    {
	s += Chain(ANY_EXCEP_ID);
    }
    else
    {
	s += Chain(OTHER_EXCEP_ID);
    }

    return s;
}
    
ostream& operator << (ostream& s, const CegoProcException& pe)
{    
    s << pe.toChain();
    return s;
}
